from typing import TYPE_CHECKING

from pyranges.core.names import END_COL, START_COL
from pyranges.core.pyranges_helpers import factorize, mypy_ensure_rangeframe

if TYPE_CHECKING:
    from pyranges.range_frame.range_frame import RangeFrame


def _merge(
    df: "RangeFrame",
    by: list[str],
    count_col: str | None = None,
    slack: int | None = None,
) -> "RangeFrame":
    import ruranges

    from pyranges.range_frame.range_frame import RangeFrame

    if df.empty:
        return df

    col_order = [col for col in df if col in [*by, START_COL, END_COL]]

    factorized = factorize(df, by)

    indices, start, end, counts = ruranges.merge(
        groups=factorized,
        starts=df.Start.to_numpy(),
        ends=df.End.to_numpy(),
        slack=slack or 0,
    )

    by_subset = df[col_order].take(indices)  # type: ignore[arg-type]
    by_subset.loc[:, START_COL] = start
    by_subset.loc[:, END_COL] = end

    outpr = RangeFrame(by_subset)

    if count_col:
        outpr.insert(outpr.shape[1], count_col, counts)

    return mypy_ensure_rangeframe(outpr.reset_index(drop=True))



import logging
from collections import OrderedDict
from typing import TYPE_CHECKING

import pandas as pd
from tabulate import tabulate

if TYPE_CHECKING:
    from pyranges import PyRanges


logging.basicConfig(level=logging.INFO)
LOGGER = logging.getLogger(__name__)
LOGGER.setLevel(logging.INFO)


def _summary(
    self: "PyRanges",
    *,
    return_df: bool = False,
) -> pd.DataFrame | None:
    lengths = {}
    total_lengths = {}
    lengths["pyrange"] = self.lengths()
    total_lengths["pyrange"] = [self.length]

    if self.strand_valid:
        c = self.merge_overlaps(use_strand=True)
        lengths["coverage_forward"] = c.loci["+"].lengths()
        lengths["coverage_reverse"] = c.loci["-"].lengths()
        total_lengths["coverage_forward"] = [c.loci["+"].length]
        total_lengths["coverage_reverse"] = [c.loci["-"].length]
    else:
        c = self

    c = c.merge_overlaps(use_strand=False)
    lengths["coverage_unstranded"] = c.lengths()
    total_lengths["coverage_unstranded"] = [c.length]

    summaries = OrderedDict()

    # stats for lengths
    for summary, s in lengths.items():
        summaries[summary] = s.describe()

    summary_lengths = pd.concat(summaries.values(), axis=1)
    summary_lengths.columns = pd.Index(summaries)

    df = pd.DataFrame.from_dict(total_lengths)
    df.index = pd.Index(["sum"])
    summary_lengths = pd.concat([summary_lengths, df])

    if return_df:
        return summary_lengths
    str_repr = tabulate(summary_lengths, headers=list(summary_lengths.columns))  # type: ignore[arg-type]
    print(str_repr)  # noqa: T201
    return None
